# -*- coding: utf-8 -*-
#
# Copyright (C) 2020-2022 Technische Universität Graz.
#
# invenio-pure is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.

"""Module tests."""

from flask import Flask

from invenio_pure import InvenioPure, __version__


def test_version() -> None:
    """Test version import."""
    assert __version__


def test_init() -> None:
    """Test extension initialization."""
    app = Flask("testapp")
    ext = InvenioPure(app)
    assert "invenio-pure" in app.extensions

    app = Flask("testapp")
    ext = InvenioPure()
    assert "invenio-pure" not in app.extensions
    ext.init_app(app)
    assert "invenio-pure" in app.extensions



# -*- coding: utf-8 -*-
#
# Copyright (C) 2020-2022 Technische Universität Graz.
#
# invenio-pure is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.

"""Pytest configuration.

See https://pytest-invenio.readthedocs.io/ for documentation on which test
fixtures are available.
"""


from collections.abc import Callable

import pytest
from flask import Flask

from invenio_pure import InvenioPure


@pytest.fixture(scope="module")
def create_app(instance_path: str) -> Callable:
    """Application factory fixture."""

    def factory(**config: dict) -> Flask:
        app = Flask("testapp", instance_path=instance_path)
        app.config.update(**config)
        InvenioPure(app)
        return app

    return factory



# -*- coding: utf-8 -*-
#
# Copyright (C) 2024 Graz University of Technology.
#
# invenio-pure is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.

"""Services."""

from .config import PureRESTServiceConfig
from .decorators import build_service
from .services import PureRESTService

__all__ = (
    "PureRESTService",
    "PureRESTServiceConfig",
    "build_service",
)



# -*- coding: utf-8 -*-
#
# Copyright (C) 2024 Graz University of Technology.
#
# invenio-pure is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.

"""API."""

from pathlib import Path
from tempfile import NamedTemporaryFile

from ..types import Filter, PureID
from .config import PureRESTConfig
from .models import PureConnection


class PureAPI:
    """Pure api."""

    connection_cls = PureConnection

    def __init__(self, config: PureRESTConfig) -> None:
        """Construct."""
        self.connection = self.connection_cls(config)

    def fetch_all_ids(self, filter_records: Filter) -> list[str]:
        """Get records."""
        records = self.connection.ids(filter_records)
        return [record["uuid"] for record in records["items"]]

    def get_metadata(self, pure_id: PureID) -> dict:
        """Get metadata."""
        return self.connection.metadata(pure_id)

    def download_file(self, file_: dict) -> str:
        """Download file."""
        filename = file_["fileName"]
        prefix = Path(filename).stem
        suffix = Path(filename).suffix

        with NamedTemporaryFile(
            delete=False,
            delete_on_close=False,
            prefix=f"{prefix}-",
            suffix=suffix,
        ) as file_pointer:
            self.connection.store_file_temporarily(file_["url"], file_pointer)
        return file_pointer.name

    def mark_as_exported(self, pure_id: PureID, record: dict) -> bool:
        """Mark as exported."""
        # TODO:
        # replace dk/atira/pure/researchoutput/keywords/export2repo/validated with
        # dk/atira/pure/researchoutput/keywords/export2repo/exported
        # not sure if the attribute worksl like that
        record["keywordUris"] = [
            "dk/atira/pure/researchoutput/keywords/export2repo/exported",
        ]
        return self.connection.mark_as_exported(pure_id, record)



# -*- coding: utf-8 -*-
#
# Copyright (C) 2022-2025 Graz University of Technology.
#
# invenio-pure is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.

"""Types of the pure connector."""


PureID = str
"""This describes the main id used in pure.

This is the reference between the invenioRDM instance and the PURE tool.
"""

URL = str
"""This type should be used wherever a URL is the value of the variable."""

PureToken = str
"""The Pure Token.

It will not have a special schema.
"""

PureRecord = dict
"""The record which will be imported from pure.

This record has to be a dict (json).
"""

FilePath = str
"""The path to the file on the local machine."""

EmailAddress = str
"""The email address.

The address does not have to have a special format, but it has to be an email.
"""

PureAPIKey = str
"""Pure api token."""

Filter = dict[str, list[str]]
"""The filter describes which records will be imported into the repository."""



# -*- coding: utf-8 -*-
#
# Copyright (C) 2024-2025 Graz University of Technology.
#
# invenio-pure is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.

"""Services decorators."""


from collections.abc import Callable
from functools import wraps
from typing import TypedDict, Unpack

from .config import PureRESTServiceConfig
from .services import PureRESTService


class KwargsDict(TypedDict, total=False):
    """Kwargs dict."""

    endpoint: str
    token: str
    user_email: str
    pure_id: str
    no_color: bool
    pure_service: PureRESTService  # not nice, try to remove that from here, mypy


def build_service[T](func: Callable[..., T]) -> Callable:
    """Decorate to build the services."""

    @wraps(func)
    def build(*_: dict, **kwargs: Unpack[KwargsDict]) -> T:
        endpoint = kwargs.pop("endpoint")
        token = kwargs.pop("token")

        config = PureRESTServiceConfig(endpoint, token)

        kwargs["pure_service"] = PureRESTService(config=config)

        return func(**kwargs)

    return build



# -*- coding: utf-8 -*-
#
# Copyright (C) 2024-2025 Graz University of Technology.
#
# invenio-pure is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.

"""Models."""


from shutil import copyfileobj
from tempfile import _TemporaryFileWrapper
from typing import cast

from requests import ReadTimeout, get, post, put
from requests.exceptions import JSONDecodeError

from ..types import URL, Filter, PureAPIKey, PureID
from .config import PureRESTConfig


class PureRESTError(Exception):
    """Pure rest error."""

    def __init__(self, code: int, msg: str) -> None:
        """Create Pure rest error."""
        super().__init__(f"Pure REST error code={code} msg='{msg}'")


class PureRESTPOSTJson:
    """Pure rest post xml."""

    def __init__(self, api_key: PureAPIKey) -> None:
        """Construct."""
        self.api_key = api_key

    def create_request_headers(self) -> dict[str, str]:
        """Headers."""
        return {
            "api-key": self.api_key,
            "accept": "application/json",
            "Content-Type": "application/json",
        }


class PureConnection:
    """Pure connection."""

    def __init__(self, config: PureRESTConfig) -> None:
        """Construct."""
        self.config = config
        self.post_json = PureRESTPOSTJson(self.config.token)

    def get(self, endpoint: str, headers: dict[str, str]) -> dict:
        """Get."""
        try:
            response = get(endpoint, headers=headers, timeout=10)
        except (ReadTimeout, JSONDecodeError) as exc:
            raise PureRESTError(code=550, msg=str(exc)) from exc

        if str(response.status_code) != "200":
            raise PureRESTError(code=response.status_code, msg=str(response.text))

        return cast(dict, response.json())

    def put(self, endpoint: str, data: dict, headers: dict[str, str]) -> dict:
        """Put."""
        try:
            response = put(endpoint, json=data, headers=headers, timeout=10)
        except (ReadTimeout, JSONDecodeError) as exc:
            raise PureRESTError(code=550, msg=str(exc)) from exc

        if str(response.status_code) != "200":
            raise PureRESTError(code=response.status_code, msg=str(response.text))

        return cast(dict, response.json())

    def post(self, endpoint: str, data: dict, headers: dict[str, str]) -> dict:
        """Post."""
        try:
            response = post(endpoint, json=data, headers=headers, timeout=10)
        except (ReadTimeout, JSONDecodeError) as exc:
            raise PureRESTError(code=550, msg=str(exc)) from exc

        if str(response.status_code) != "200":
            raise PureRESTError(code=response.status_code, msg=str(response.text))

        return cast(dict, response.json())

    def ids(self, filter_records: Filter) -> dict:
        """Post ids."""
        endpoint = f"{self.config.endpoint}/search"
        body = filter_records
        headers = self.post_json.create_request_headers()

        return self.post(endpoint, body, headers)

    def metadata(self, pure_id: PureID) -> dict:
        """Post metadata."""
        endpoint = f"{self.config.endpoint}/{pure_id}"
        headers = self.post_json.create_request_headers()

        return self.get(endpoint, headers)

    def store_file_temporarily(
        self,
        file_url: URL,
        file_pointer: _TemporaryFileWrapper,
    ) -> None:
        """Download a file from Pure to given destination path with given file name.

        Return path to the downloaded file upon success, empty string upon failure.
        """
        headers = {"api-key": self.config.token}
        with get(file_url, stream=True, headers=headers, timeout=10) as response:
            copyfileobj(response.raw, file_pointer)

    def mark_as_exported(self, pure_id: PureID, record: dict) -> bool:
        """Post mark as exported."""
        endpoint = f"{self.config.endpoint}/{pure_id}"
        body = record
        headers = self.post_json.create_request_headers()
        # TODO check return value, json may contain error message and return False then
        self.put(endpoint, body, headers)
        return True



# -*- coding: utf-8 -*-
#
# Copyright (C) 2021-2024 Graz University of Technology.
#
# invenio-pure is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.

"""Scheduled tasks for celery."""


from celery import shared_task
from flask import current_app
from invenio_access.permissions import system_identity

from .proxies import current_pure


@shared_task(ignore_result=True)
def import_records_from_pure() -> None:
    """Import records from pure."""
    import_func = current_app.config["PURE_IMPORT_FUNC"]
    filter_records = current_app.config["PURE_FILTER_RECORDS"]

    pure_service = current_pure.pure_rest_service
    ids = pure_service.fetch_all_ids(system_identity, filter_records)

    for pure_id in ids:
        try:
            import_func(system_identity, pure_id, pure_service)
        except RuntimeError as e:
            msg = "ERROR pure pure_id: %s with message: %s"
            current_app.logger.error(msg, pure_id, str(e))



# -*- coding: utf-8 -*-
#
# Copyright (C) 2024-2025 Graz University of Technology.
#
# invenio-pure is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.

"""Services."""


from flask_principal import Identity

from ..records import PureAPI
from ..types import Filter, PureID
from .config import PureRESTServiceConfig


class PureRESTService:
    """Pure rest service."""

    def __init__(self, config: PureRESTServiceConfig) -> None:
        """Construct."""
        self._config = config
        self.api = self.api_cls(config=config)

    @property
    def api_cls(self) -> type[PureAPI]:
        """Get api cls."""
        return self._config.api_cls

    def fetch_all_ids(self, _: Identity, filter_records: Filter) -> list[str]:
        """Fetch all ids from pure by given filter."""
        return self.api.fetch_all_ids(filter_records)

    def get_metadata(self, _: Identity, pure_id: PureID) -> dict:
        """Get metadata by pure uuid."""
        return self.api.get_metadata(pure_id)

    def download_file(self, _: Identity, file_: dict) -> str:
        """Download File."""
        return self.api.download_file(file_)

    def mark_as_exported(self, _: Identity, pure_id: PureID, record: dict) -> bool:
        """Mark as exported."""
        return self.api.mark_as_exported(pure_id, record)



# -*- coding: utf-8 -*-
#
# Copyright (C) 2021-2025 Graz University of Technology.
#
# invenio-pure is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.

"""Invenio module that adds pure."""

from flask import Flask

from .services import PureRESTService, PureRESTServiceConfig


class InvenioPure:
    """invenio-pure extension."""

    def __init__(self, app: Flask | None = None) -> None:
        """Extension initialization."""
        if app:
            self.init_app(app)

    def init_app(self, app: Flask) -> None:
        """Flask application initialization."""
        self.init_services(app)
        app.extensions["invenio-pure"] = self

    def init_services(self, app: Flask) -> None:
        """Initialize services."""
        endpoint = app.config.get("PURE_ENDPOINT", "")
        token = app.config.get("PURE_TOKEN", "")
        config = PureRESTServiceConfig(endpoint, token)
        self.pure_rest_service = PureRESTService(config)



# -*- coding: utf-8 -*-
#
# Copyright (C) 2024 Graz University of Technology.
#
# invenio-pure is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.

"""Helper proxy to the state object."""

from typing import cast

from flask import current_app
from werkzeug.local import LocalProxy

from .ext import InvenioPure

current_pure: InvenioPure = cast(
    InvenioPure,
    LocalProxy(
        lambda: current_app.extensions["invenio-pure"],
    ),
)
"""Helper proxy to get teh current pure extension."""



# -*- coding: utf-8 -*-
#
# Copyright (C) 2024 Graz University of Technology.
#
# invenio-pure is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.

"""Services configs."""

from dataclasses import dataclass

from ..records import PureAPI, PureRESTConfig


@dataclass
class PureRESTServiceConfig(PureRESTConfig):
    """Pure REST service config."""

    api_cls: type[PureAPI] = PureAPI



# -*- coding: utf-8 -*-
#
# Copyright (C) 2024-2025 Graz University of Technology.
#
# invenio-pure is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.

"""REST config."""


from dataclasses import dataclass

from ..types import URL, PureToken


@dataclass
class PureRESTConfig:
    """Pure rest config."""

    endpoint: URL = ""
    token: PureToken = ""



# -*- coding: utf-8 -*-
#
# Copyright (C) 2024 Graz University of Technology.
#
# invenio-pure is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.

"""Records."""

from .api import PureAPI
from .config import PureRESTConfig
from .models import PureConnection

__all__ = (
    "PureAPI",
    "PureConnection",
    "PureRESTConfig",
)



# -*- coding: utf-8 -*-
#
# Copyright (C) 2021-2025 Graz University of Technology.
#
# invenio-pure is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.

"""Invenio module that adds pure."""


from .types import URL, EmailAddress, PureToken

PURE_CELERY_BEAT_SCHEDULE: dict[str, dict] = {}
"""The celery beat schedule is used to configure the import schedule.

Following is a example configuration. One option to add this to the
global celery beat schedule is to update the CELERY_BEAT_SCHEDULE
dict in invenio.cfg by the example below:

.. code-block:: python

    "synchronize_records": {
        "task": "invenio_pure.tasks.import_records_from_pure",
        "schedule": crontab(hour=1, minute=0, day_of_week=0),
    }
"""

PURE_PURE_ENDPOINT: URL = ""
"""This is the endpoint of the pure instance."""

PURE_PURE_TOKEN: PureToken = ""
"""This is the token to be allowed to use the API."""

PURE_USER_EMAIL: EmailAddress = ""
"""This is the user email of the pure user within the InvenioRDM instance."""



# -*- coding: utf-8 -*-
#
# Copyright (C) 2021-2024 Graz University of Technology.
#
# invenio-pure is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.

"""Invenio module that adds pure."""

from .errors import PureAPIError, PureRuntimeError
from .ext import InvenioPure
from .types import URL, PureID

__version__ = "1.0.0"

__all__ = (
    "__version__",
    "InvenioPure",
    "PureID",
    "PureAPIError",
    "PureRuntimeError",
    "URL",
)



# -*- coding: utf-8 -*-
#
# Copyright (C) 2022-2023 Technische Universität Graz
#
# invenio-pure is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.

"""REST errors."""

from .types import PureRecord


class PureRuntimeError(RuntimeError):
    """Pure runtime error."""

    def __init__(self, record: PureRecord) -> None:
        """Construct for pure runtime error."""
        self.record = record


class PureAPIError(Exception):
    """Pure API error class."""

    def __init__(self, code: int, msg: str) -> None:
        """Construct for pure api error."""
        super().__init__(f"Pure API error code={code} msg='{msg}'")



# -*- coding: utf-8 -*-
#
# Copyright (C) 2021-2025 Graz University of Technology.
#
# invenio-pure is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.

"""CLI commands for Invenio-RDM-Pure."""


from click import STRING, group, option, secho
from click_params import URL
from flask import current_app
from flask.cli import with_appcontext
from invenio_access.utils import get_identity
from invenio_accounts import current_accounts

from .services import PureRESTService, build_service


@group()
def pure() -> None:
    """Commands for InvenioRdmPure."""


@pure.command("import")
@with_appcontext
@option("--endpoint", type=URL, required=True)
@option("--token", type=STRING, required=True)
@option("--user-email", type=STRING, required=True)
@option("--pure-id", type=STRING, required=True)
@option("--no-color", is_flag=True, default=False)
@build_service
def import_records_from_pure(
    pure_service: PureRESTService,
    user_email: str,
    pure_id: str,
    *,
    no_color: bool,
) -> None:
    """Import a record from given JSON file or JSON string."""
    import_func = current_app.config["PURE_IMPORT_FUNC"]
    user = current_accounts.datastore.get_user_by_email(user_email)
    identity = get_identity(user)
    try:
        record = import_func(identity, pure_id, pure_service)
        color = "green" if not no_color else "black"
        secho(f"record.id: {record.id}", fg=color)
    except RuntimeError as e:
        msg = f"ERROR pure pure_id: {pure_id} with message: {e!r}"
        color = "red" if not no_color else "black"
        secho(msg, fg=color)


@pure.command("list")
@with_appcontext
@option("--endpoint", type=URL, required=True)
@option("--token", type=STRING, required=True)
@option("--user-email", type=STRING, required=True)
@build_service
def list_all_available_records(pure_service: PureRESTService, user_email: str) -> None:
    """List all possible to import records."""
    filter_records = current_app.config["PURE_FILTER_RECORDS"]
    user = current_accounts.datastore.get_user_by_email(user_email)
    identity = get_identity(user)

    for pure_id in pure_service.fetch_all_ids(identity, filter_records):
        secho(f"pure_id: {pure_id}", fg="green")


@pure.command("sync")
@with_appcontext
@option("--endpoint", type=URL, required=True)
@option("--token", type=STRING, required=True)
@option("--user-email", type=STRING, required=True)
@build_service
def sync(pure_service: PureRESTService, user_email: str) -> None:
    """Sync Pure with the repo."""
    import_func = current_app.config["PURE_IMPORT_FUNC"]
    filter_records = current_app.config["PURE_FILTER_RECORDS"]
    user = current_accounts.datastore.get_user_by_email(user_email)
    identity = get_identity(user)

    ids = pure_service.fetch_all_ids(identity, filter_records)

    for pure_id in ids:
        try:
            import_func(identity, pure_id, pure_service)
            msg = f"SUCCESS pure pure_id: {pure_id} imported successfully."
            secho(msg, fg="green")
        except RuntimeError as e:
            msg = f"ERROR pure pure_id: {pure_id} with message: {e!r}"
            secho(msg, fg="red")
